/*
    main.c

    Copyright © 2014 thinwire

    This file is part of rigoltmc.

    rigoltmc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rigoltmc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rigoltmc.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <iostream>
#include <libusb-1.0/libusb.h>
#include "config.h"
#include "usbtmcsimple.h"


using namespace std;

int optionerror (void)
{

   cerr<<"Option has invalid format"<<endl;
   return 1;
}


int do_command_with_restore (
      int verbose,
      const char *verbosetext,
      const char *command,
      const char *datfile,
      int vid,
      int pid)
{
   libusb_device_handle *dev_handle; //a device handle
   int ret;
   unsigned char bTag = 1;
   char cmd[4096];
   unsigned char setup_data[4096];
   int actual;
   int size;

   dev_handle = usbtmc_init (verbose, vid, pid);
   if (dev_handle == NULL)
      return 1;

   /* Request for system setup data */
   if (verbose)
      cout<<"Save System Setup data."<<endl;

   strcpy(cmd, ":SYSTem:SETup?");
   ret = usbtmc_write(dev_handle, strlen(cmd), cmd, bTag);
   if (!ret)
   {
      usbtmc_shutdown(dev_handle);
      return 1;
   }
	
   /* Read system setup data */
   actual = sizeof(setup_data);
   ret = usbtmc_read(dev_handle, actual, setup_data, bTag);
   if (!ret)
   {
      usbtmc_shutdown(dev_handle);
      return 1;
   }

   if (datfile != NULL)
   {
      FILE *fp;

      if (verbose)
         cout<<"Writing setup data to "<<datfile<<"."<<endl;

      if ((fp = fopen(datfile, "wb")) != NULL)
      {
         fwrite(setup_data, actual, 1, fp);
         fclose(fp);
      } else {
         if (verbose)
            cout<<"*** Error opening file "<<datfile<<" for writing."<<endl;
      }
   }

   if (verbose)
      cout<<verbosetext<<endl;

   ret = usbtmc_write(dev_handle, strlen(command), command, bTag);

   if (verbose)
      cout<<"Restore System Setup data."<<endl;

   strcpy(cmd, ":SYST:SET ");
   size = strlen(cmd);
   memcpy(&cmd[size], setup_data, actual);
   size += actual;
   ret = usbtmc_write(dev_handle, size, cmd, bTag);

   usbtmc_shutdown(dev_handle);

   return 0;
}

int install_option (
       int verbose,
       const char *poption,
       const char *datfile,
       int vid,
       int pid)
{
   char verbosetext[64];
   char cmd[32];
   char option[29];
   const char *p = poption;
   unsigned int i;

   // Remove '-' characters from option string
   for (i = 0; i < sizeof(option) - 1; i++)
   {
      while (*p == '-')
         p++;

      if (*p == '\0')
         return optionerror(); // Option string too short
      option[i] = *(p++);
   }
   option[i] = '\0';

   if (*p != '\0')
      return optionerror(); // Option string too long
   
   strcpy(verbosetext, "Install Option: ");
   strcat(verbosetext, poption);
   strcpy(cmd, ":SYSTem:OPTion:INSTall ");
   strcat(cmd, option);

   return do_command_with_restore (
             verbose,
             verbosetext,
             cmd,
             datfile,
             vid,
             pid);
}

int uninstall_all_options (
       int verbose, 
       const char *datfile,
       int vid,
       int pid)
{
   const char *verbosetext = "Uninstall all options.";
   const char *cmd         = ":SYSTem:OPTion:UNINSTall";

   return do_command_with_restore (
             verbose,
             verbosetext,
             cmd,
             datfile,
             vid,
             pid);
}

int get_instrument_id (int verbose, int vid, int pid)
{
   libusb_device_handle *dev_handle;
   char cmd[64];
   unsigned char data[500];
   unsigned char bTag = 1;
   int actual;
   int ret;

   dev_handle = usbtmc_init(verbose, vid, pid);
   if (dev_handle == NULL)
      return 1;

   if (verbose)
      cout<<"Query instrument ID"<<endl;

   strcpy(cmd, "*IDN?");
   usbtmc_write(dev_handle, strlen(cmd), cmd, bTag);

   actual = sizeof(data);
   ret = usbtmc_read(dev_handle, actual, data, bTag);
   if (!ret)
   {
      usbtmc_shutdown(dev_handle);
      return 1;
   }

   data[actual] = '\0';
   cout<<data<<endl;

   usbtmc_shutdown(dev_handle);

   return 0;
}

int help(void)
{
   cout<<"rigoltmc V" VERSION<<"."<<endl;
   cout<<"Copyright (C) 2014 thinwire"<<endl;
   cout<<endl;
   cout<<"Syntax:"<<endl;
   cout<<"   rigoltmc (-h|-i|-u|-o<option>)[-v]"<<endl;
   cout<<"Parameters:"<<endl;
   cout<<"   -h         print command syntax"<<endl;
   cout<<"   -i         get instrument ID"<<endl;
   cout<<"   -o<option> install option"<<endl;
   cout<<"   -u         uninstall all options"<<endl;
   cout<<"   -v         verbose output"<<endl;
   cout<<"Example:"<<endl;
   cout<<"   rigoltmc -v -iAFEH2H1-FEI89CH-CL6WEIC-CS43CZY"<<endl;

   return 0;
}

enum eCommand {
   GETINSTID,
   UNINSTALL,
   INSTALL,
   HELP
};

int main (int argc, char **argv)
{
   const char *option  = NULL;
   const char *datfile = NULL;
   eCommand cmd = HELP;
   int valid = 1;
   int c;
   int vid = RIGOL_USB_VID; // Rigols Vendor ID
   //int pid = RIGOL_USB_DS1000_PID;
   int pid = RIGOL_USB_DS2000_PID; // Rigol DSO DS2000 Series Product ID
   //int pid = RIGOL_USB_DS4000_PID;
   int verbose = 0;
 
   opterr = 0;

   while ((c = getopt (argc, argv, "vhiuo:")) != -1)
   {
      switch (c)
      {
         case 'v':
            verbose = 1;
            break;
         case 'i':
            if (cmd != HELP) { valid = 0; };
            cmd = GETINSTID;
            break;
         case 'o':
            if (cmd != HELP) { valid = 0; };
            cmd = INSTALL;
            option = optarg;
            break;
         case 'u':
            if (cmd != HELP) { valid = 0; };
            cmd = UNINSTALL;
            break;
         case 'h':
            if (cmd != HELP) { valid = 0; };
            cmd = HELP;
            break;
         case '?':
            if (optopt == 'i')
               cerr<<"Option '-"<<(char)optopt<<"' requires an argument."<<endl;
            else if (isprint (optopt))
               cerr<<"Unknown option '-"<<(unsigned char)optopt<<"'."<<endl;
            else {
               cerr.width(2);
               cerr<<"Unknown option character with code 0x"<<std::hex<<
                      optopt<<"."<<endl;
            }
            return 1;
         default:
            break;
      }
   }

   if (!valid)
   {
      cerr<<"*** Error: Invalid combination of command line parameters"<<endl;
      cmd = HELP;
   }

   switch (cmd)
   {
      case GETINSTID: return get_instrument_id (verbose, vid, pid);
      case INSTALL:   return install_option (verbose, option, datfile, vid, pid);
      case UNINSTALL: return uninstall_all_options (verbose, datfile, vid, pid);
      case HELP:      return help();
   }

   return 0;
}

