/*
    usbtmcsimple.c

    Copyright © 2014 thinwire

    This file is part of rigoltmc.

    rigoltmc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    rigoltmc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with rigoltmc.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string.h>
#include <libusb-1.0/libusb.h>
#include "usbtmcsimple.h"


using namespace std;

static libusb_context *ctx = NULL; //a libusb session
unsigned char *io_buffer = NULL;
int verbose  = 0;
int detached = 0;

libusb_device_handle *usbtmc_init(int bVerbose, int vid, int pid)
{
        libusb_device **devs; //pointer to pointer of device, used to retrieve a list of devices
        libusb_device_handle *dev_handle = NULL; //a device handle
	ssize_t cnt; //holding number of devices in list
	unsigned char desc[128];
	int ret;

        ret = libusb_init(&ctx);
        if(ret < 0)
        {
           cout<<"*** Error initializing libusb. Error code: "<<ret<<"."<<endl;
           return NULL;
        }
        libusb_set_debug(ctx, 3); // set verbosity level to 3,
                                  // as suggested in the documentation

        cnt = libusb_get_device_list(ctx, &devs); //get the list of devices
        if(cnt < 0) {
           cout<<"*** Error getting USB Devices."<<endl;
           usbtmc_shutdown(dev_handle);
           return NULL;
        }

	if (bVerbose)
           cout<<cnt<<" USB Devices detected."<<endl;

        dev_handle = libusb_open_device_with_vid_pid(ctx, vid, pid);
        libusb_free_device_list(devs, 1); //free the list, unref the devices in it
        if(dev_handle == NULL)
        {
           cout<<"*** Error opening USB Rigol device."<<endl;
           usbtmc_shutdown(dev_handle);
           return NULL;
        } else
	   if (bVerbose)
           {
              cout<<"USB Device ";
              if (libusb_get_string_descriptor_ascii(dev_handle, 1, desc, 128) > 0)
                 cout<<desc;
              if (libusb_get_string_descriptor_ascii(dev_handle, 2, desc, 128) > 0)
                 cout<<" "<<desc;
              cout<<" opened."<<endl;
           }

        if (libusb_kernel_driver_active(dev_handle, 0) == 1)
        { //find out if kernel driver is attached
	   if (bVerbose)
              cout<<"Kernel Driver Active."<<endl;
           if(libusb_detach_kernel_driver(dev_handle, 0) == 0)
           {
              detached = 1;
              // Sucessfully detach it
	      if (bVerbose)
                 cout<<"Kernel Driver Detached"<<endl;
           }
        }
        //claim interface 0 (the first) of device
        ret = libusb_claim_interface(dev_handle, 0);

        if (ret < 0)
        {
           cout<<"*** Error claiming USB-Interface."<<endl;
           usbtmc_shutdown(dev_handle);
           return NULL;
        }

	if (bVerbose)
           cout<<"Claimed USB-Interface."<<endl;

        verbose = bVerbose;
	io_buffer = (unsigned char *)malloc(USBTMC_SIZE_IOBUFFER);

	return dev_handle;
}

int usbtmc_write(libusb_device_handle *dev_handle,
	  int cmdsize,
          const char *cmd,
          unsigned char &bTag)
{
   int isrigol = 0;
   int remain;
   int headersize = USBTMC_SIZE_HEADER;
   int msgsize = cmdsize;
   int ret;
   libusb_device_descriptor dev_desc;

   remain = cmdsize;

   ret = libusb_get_descriptor(
            dev_handle,
            LIBUSB_DT_DEVICE,
            0,
            (unsigned char *)&dev_desc,
            sizeof(dev_desc));


   /*
      The Rigol DS2000 DSO series device only allows to send a
      USBTMC command message with one transfer (Containing one
      USB bulk OUT header). I tried with multiple transfers and
      got sproadic LIBUSB_ERROR_TIMEOUT return codes.
      Maybe this could be verified by somebody else.
      In this case the header contains the all-over message size.
      The flag 'isrigol' controls this Rigol specific limitation.
   */ 
   if (ret == sizeof(dev_desc) && (dev_desc.idVendor == RIGOL_USB_VID))
   {
      isrigol = 1;
      if (verbose)
         cout<<"Detected a Rigol device"<<endl;
   }

   while (remain > 0)
   {
      int islast;
      int actual;
      int size;
      int bytes;

      if (remain > USBTMC_SIZE_IOBUFFER - headersize)
      {
         size = USBTMC_SIZE_IOBUFFER - headersize;
	 islast  = 0; // Not last message

      } else {
         size = remain;
	 islast  = 0x01; // Last message
      }

      if (!isrigol)
         msgsize = size;

	/* Create a Bulk-OUT USBTMC command message header */
      if (headersize > 0)
      {
         io_buffer[0]  = DEV_DEP_MSG_OUT;
         io_buffer[1]  = bTag;
         io_buffer[2]  = ~bTag;
         io_buffer[3]  = 0; // Reserved
         io_buffer[4]  = msgsize & 0xff;
         io_buffer[5]  = (msgsize >> 8) & 0xff;
         io_buffer[6]  = (msgsize >> 16) & 0xff;
         io_buffer[7]  = (msgsize >> 24) & 0xff;
         io_buffer[8]  = isrigol ? 0x01 : islast;
         io_buffer[9]  = 0; // Reserved
         io_buffer[10] = 0; // Reserved
         io_buffer[11] = 0; // Reserved
      }

      memcpy(&io_buffer[headersize], cmd+(cmdsize-remain), size);
      bytes = headersize + size;
      if (bytes % 4)
      {
        int temp = 4 - (bytes % 4);
        memset(&io_buffer[bytes], 0, temp);
	bytes += temp;
      }

      if (verbose)
         cout<<"Writing Data..."<<endl;

      ret = libusb_bulk_transfer(
             dev_handle,
             USB_BULKOUT_ADDRESS | LIBUSB_ENDPOINT_OUT,
             io_buffer,
             bytes,
             &actual,
             USBTMC_TIMEOUT);

      if (headersize > 0)
	 if (++bTag == 0)
	    bTag++;
      if (isrigol)
         headersize = 0;

	if (ret < 0)
        {
           cout<<"*** Write Error. Error code: "<<ret<<"."<<endl;
              return 0;
	}

        if (verbose)
           cout<<"Writing Successful! Written "<<size<<" bytes."<<endl;

        remain -= size;
   }

   return 1;
}

int usbtmc_read(libusb_device_handle *dev_handle,
	  int &rsize,
          unsigned char *rdata,
          unsigned char &bTag)
{
   int sum = 0;

   do {
      int bytes;
      int size;
      int actual;
      int ret;

      if (rsize > USBTMC_SIZE_IOBUFFER - USBTMC_SIZE_HEADER - 3)
         size = USBTMC_SIZE_IOBUFFER - USBTMC_SIZE_HEADER - 3;
      else
         size = rsize;

      /* Create a Bulk-IN USBTMC command message header */
      io_buffer[0]  = REQUEST_DEV_DEP_MSG_IN;
      io_buffer[1]  = bTag;
      io_buffer[2]  = ~bTag;
      io_buffer[3]  = 0; // Reserved
      io_buffer[4]  = size & 0xff;
      io_buffer[5]  = (size >> 8) & 0xff;
      io_buffer[6]  = (size >> 16) & 0xff;
      io_buffer[7]  = (size >> 24) & 0xff;
      io_buffer[8]  = 0; // Ignore TermChar
      io_buffer[9]  = 0; // TermChar (ignored)
      io_buffer[10] = 0; // Reserved
      io_buffer[11] = 0; // Reserved

      if (verbose)
         cout<<"Writing Read request..."<<endl;

      ret = libusb_bulk_transfer(
         dev_handle,
         USB_BULKOUT_ADDRESS | LIBUSB_ENDPOINT_OUT,
         io_buffer,
         USBTMC_SIZE_HEADER,
         &actual,
         USBTMC_TIMEOUT);

      if (++bTag == 0)
         bTag++;

      if(ret < 0)
      {
         cout<<"*** Write Error. Error code: "<<ret<<"."<<endl;
         return 0;
      }

      if (verbose)
         cout<<"Writing Successful!"<<endl;


      ret = libusb_bulk_transfer(
         dev_handle,
         USB_BULKIN_ADDRESS | LIBUSB_ENDPOINT_IN,
         io_buffer,
         USBTMC_SIZE_IOBUFFER,
         &actual,
         USBTMC_TIMEOUT);

      if(ret < 0)
      {
         cout<<"*** Read Error. Error code: "<<ret<<"."<<endl;
         return 0;
      }   

      bytes = io_buffer[4] +
	(io_buffer[5] <<  8) +
	(io_buffer[6] << 16) +
   	(io_buffer[7] << 24);

      if (verbose)
         cout<<"Reading Successful! Read "<<bytes<<" bytes."<<endl;
      if (0)
      {
         cout<<"Data junk:"<<endl;
         for (int i = 0; i < actual; i++)
         {		
            cout<<" 0x"<<hex<<(unsigned int)io_buffer[i];
            if (!((i+1) % 16))
              cout<<endl;
         }
         cout<<endl;
      }

      memcpy(rdata+sum, &io_buffer[USBTMC_SIZE_HEADER], bytes);
      sum += bytes;
   } while (!(io_buffer[8] & 0x01)); // Last data junk read => done

   if (verbose)
      cout<<"Sum of bytes read: "<<sum<<"."<<endl;
   rsize = sum; // return sum of bytes read

   return 1;
}

void usbtmc_shutdown(libusb_device_handle *dev_handle)
{
   if (io_buffer != NULL)
   {
      free(io_buffer);
      io_buffer = NULL;
   }

   if (dev_handle != NULL)
   {
      int ret;

      ret = libusb_release_interface(dev_handle, 0);
      if (verbose)
      {
         if (ret < 0)
            cout<<"*** Error releasing USB-Interface."<<endl;
         else 
            cout<<"Released USB-Interface."<<endl;
      }

      if(detached && libusb_attach_kernel_driver(dev_handle, 0) == 0)
      {
         // Sucessfully attached
         if (verbose)
             cout<<"Kernel Driver Attached"<<endl;
      }

      libusb_close(dev_handle);
      dev_handle = NULL;
   }
   detached = 0;

   if (ctx != NULL)
   {
      libusb_exit(ctx);
      ctx = NULL;
   }
}

